import datetime
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union

from .deprecation_utils import deprecated_method
from .deprecation_utils import warn_deprecated_module
from .types import ArchiveStatusLevel
from .types import ArchiveStatusLevel as StatusLevel  # noqa F401 deprecated
from .types import ArchiveStatusType
from .types import ArchiveStatusType as StatusType  # noqa F401 deprecated
from .types import Dataset
from .types import DatasetId
from .types import DatasetMetadata  # noqa F401
from .types import SampleMetadata

warn_deprecated_module()


class IcatClientInterface:
    def send_message(
        self,
        msg: str,
        msg_type: Optional[str] = None,
        beamline: Optional[str] = None,
        proposal: Optional[str] = None,
        investigation_id: Optional[str] = None,
        dataset: Optional[str] = None,
        beamline_only: Optional[bool] = None,
        **payload,
    ):
        """
        Send a message to the proposal or beamline e-logbook.

        :param msg: The message content.
        :param msg_type: {'comment', 'debug', 'info', 'error', 'commandLine'}, optional.
        :param beamline: The beamline name of the proposal or beamline e-logbook.
        :param proposal: The proposal name of the e-logbook. Ignored if `beamline_only` is True.
        :param investigation_id: The investigation identifier in the ICAT database.
        :param dataset: The specific dataset name to link the message to.
        :param beamline_only: if `True`, the message will be stored in the beamline e-logbook.
        :param editable: Used with the `formatted` field, to determine the category of message. Annotation characterizes editable and unformatted messages, while Notification charaterizes non-editable and formatted messages.
        :param formatted: Used with the `editable` field, to determine the category of message. Annotation characterizes editable and unformatted messages, while Notification charaterizes non-editable and formatted messages.
        :param mimetype: {'text/plain', 'text/html'}, optional.
        :param payload: Additional payload for the message. It can contain tags (list of strings or list of dictionaries), the machine, the software.
        """
        raise NotImplementedError

    def disconnect(self):
        pass

    def send_binary_data(
        self,
        data: bytes,
        mimetype: Optional[str] = None,
        beamline: Optional[str] = None,
        proposal: Optional[str] = None,
        investigation_id: Optional[str] = None,
        beamline_only: Optional[bool] = None,
        **payload,
    ):
        """
        Send an image in base64 format to the proposal or beamline e-logbook.

        :param data: The binary message content.
        :param mimetype: {'text/plain', 'text/html'}, optional.
        :param beamline: The beamline name of the proposal or beamline e-logbook.
        :param proposal: The proposal name of the e-logbook. Ignored if `beamline_only` is True.
        :param investigation_id: The investigation identifier in the ICAT database.
        :param beamline_only: if `True`, the message will be stored in the beamline e-logbook.
        :param payload: Additional payload for the e-logbook message. It can contain tags (list of strings or list of dictionaries), the machine, the software.
        """
        raise NotImplementedError

    def send_text_file(
        self,
        filename: str,
        beamline: Optional[str] = None,
        proposal: Optional[str] = None,
        dataset: Optional[str] = None,
        investigation_id: Optional[str] = None,
        beamline_only: Optional[bool] = None,
        **payload,
    ):
        """
        Send the content of a text file as a message to the proposal or beamline e-logbook.

        :param filename: The filename containing the message to be sent.
        :param beamline: The beamline name of the proposal or beamline e-logbook.
        :param proposal: The proposal name of the e-logbook. Ignored if `beamline_only` is True.
        :param investigation_id: The investigation identifier in the ICAT database.
        :param beamline_only: if `True`, the message will be stored in the beamline e-logbook.
        :param payload: Additional payload for the e-logbook message. It can contain tags (list of strings or list of dictionaries), the machine, the software.
        """
        raise NotImplementedError

    def send_binary_file(
        self,
        filename: str,
        beamline: Optional[str] = None,
        proposal: Optional[str] = None,
        beamline_only: Optional[bool] = None,
        **payload,
    ):
        """
        Send the content of a file as a binary image to the proposal or beamline e-logbook.

        :param filename: The filename of the image to be sent.
        :param beamline: The beamline name of the proposal or beamline e-logbook.
        :param proposal: The proposal name of the e-logbook. Ignored if `beamline_only` is True.
        :param beamline_only: if `True`, the message will be stored in the beamline e-logbook.
        :param payload: Additional payload for the e-logbook message. It can contain tags (list of strings or list of dictionaries), the machine, the software.
        """
        raise NotImplementedError

    def start_investigation(
        self,
        beamline: Optional[str] = None,
        proposal: Optional[str] = None,
        start_datetime=None,
        end_datetime=None,
    ):
        """
        Send a message to ActiveMQ to either synchronize the experiment session from the User Portal in ICAT or to create a test experiment session in ICAT.

        :param beamline: The beamline name of the proposal.
        :param proposal: The proposal name.
        :param start_datetime: The start date of the experiment session, timezone local time. Current date time by default.
        :param end_datetime: The end date of the experiment session, timezone local time.
        """
        raise NotImplementedError

    def store_dataset(
        self,
        beamline: Optional[str] = None,
        proposal: Optional[str] = None,
        dataset: Optional[str] = None,
        path: Optional[str] = None,
        metadata: dict = None,
        store_filename: Optional[str] = None,
    ):
        """
        Request icat to store raw dataset.

        :param beamline str: beamline name like id01, id15a, bm18...
        :param proposal str: proposal name like in1169, blc14795, ihma429...
        :param str dataset: dataset name.
        :param str path: path to the raw dataset to store. Must be a folder.
        :param dict metadata: metadata to associate with the dataset. Must contain keys defined by the appropriate application definition, available at https://icat-esrf-definitions.readthedocs.io.
        :param str store_filename: xml file with metadata to be stored.
        """

        raise NotImplementedError

    def store_processed_data(
        self,
        beamline: Optional[str] = None,
        proposal: Optional[str] = None,
        dataset: Optional[str] = None,
        path: Optional[str] = None,
        metadata: dict = None,
        raw: Sequence = tuple(),
        store_filename: Optional[str] = None,
    ):
        """
        Request icat to store a processed dataset.

        :param beamline str: beamline name like id01, id15a, bm18...
        :param proposal str: proposal name like in1169, blc14795, ihma429...
        :param str dataset: dataset name like sample_XYZ.
        :param str path: path to the processed dataset to store. Can be a file or a folder.
        :param dict metadata: metadata to associate with the dataset. Must contain keys defined by the appropriate application definition, available at https://icat-esrf-definitions.readthedocs.io.
        :param tuple raw: Path(s) to the raw dataset(s). Should point to the 'bliss dataset' folder(s).
                          If processing relies on multiple datasets, all corresponding folders must be provided.
        :param str store_filename: xml file with metadata to be stored.
        """
        raise NotImplementedError

    def store_dataset_from_file(self, store_filename: Optional[str] = None):
        """
        Send a message to ActiveMQ to store a dataset and the associated metadata from a xml file stored on the disk.

        :param store_filename: The XML filename containing all dataset metadata.
        """
        raise NotImplementedError

    def investigation_info(
        self,
        beamline: str,
        proposal: str,
        date: Optional[Union[datetime.datetime, datetime.date]] = None,
        allow_open_ended: bool = True,
        timeout: Optional[float] = None,
    ) -> Optional[dict]:
        """
        Return the information of the experiment session corresponding to a beamline, proposal and date.

        :param beamline: The beamline name.
        :param proposal: The proposal name.
        :param date: The date of the proposal, current date by default.
        :param allow_open_ended: If `True`, enable to select an unofficial experiment session.
        :param timeout: Set a timeout for the ICAT request.
        :returns: If found, return the proposal, beamline, e-logbbok url and data portal url of the experiment session.
        """
        raise NotImplementedError

    def registered_dataset_ids(
        self,
        beamline: str,
        proposal: str,
        date: Optional[Union[datetime.datetime, datetime.date]] = None,
        allow_open_ended: bool = True,
        timeout: Optional[float] = None,
    ) -> Optional[List[DatasetId]]:
        """
        Return the dataset list of an experiment session.

        :param beamline: The beamline name of the proposal.
        :param proposal: The proposal name.
        :param date: The date of the proposal, current date by default.
        :param allow_open_ended: If `True`, enable to select an unofficial experiment session.
        :param timeout: Set a timeout for the ICAT request.
        :returns: The list of datasets (name and path).
        """
        raise NotImplementedError

    def registered_datasets(
        self,
        beamline: str,
        proposal: str,
        date: Optional[Union[datetime.datetime, datetime.date]] = None,
        allow_open_ended: bool = True,
        timeout: Optional[float] = None,
    ) -> Optional[List[Dataset]]:
        """
        Return the dataset information list of an experiment session.

        :param beamline: The beamline name of the proposal.
        :param proposal: The proposal name.
        :param date: The date of the proposal, current date by default.
        :param allow_open_ended: If `True`, enable to select an unofficial experiment session.
        :param timeout: Set a timeout for the ICAT request.
        :returns: The list of datasets (name, path, ICAT identifier, and :class:`.DatasetMetadata`).
        """
        raise NotImplementedError

    def investigation_info_string(
        self,
        beamline: str,
        proposal: str,
        date: Optional[Union[datetime.datetime, datetime.date]] = None,
        allow_open_ended: bool = True,
        timeout: Optional[float] = None,
    ) -> str:
        """
        Return the experiment session information as a string.

        :param beamline: The beamline name.
        :param proposal: The proposal name.
        :param date: The date of the proposal, current date by default.
        :param allow_open_ended: If `True`, enable to select an unofficial experiment session.
        :param timeout: Set a timeout for the ICAT request.
        :returns: If found, return the experiment session information from the metadata catalog as a string.
        """
        raise NotImplementedError

    def investigation_summary(
        self,
        beamline: str,
        proposal: str,
        date: Optional[Union[datetime.datetime, datetime.date]] = None,
        allow_open_ended: bool = True,
        timeout: Optional[float] = None,
    ) -> List[Tuple]:
        """
        Return the experiment session information as a `Tuple`.

        :param beamline: The beamline name.
        :param proposal: The proposal name.
        :param date: The date of the proposal, current date by default.
        :param allow_open_ended: If `True`, enable to select an unofficial experiment session.
        :param timeout: Set a timeout for the ICAT request.
        :returns: If found, return the experiment session information from the metadata catalog as a `Tuple`.
        """
        raise NotImplementedError

    @deprecated_method
    def update_archive_restore_status(
        self,
        dataset_id: int = None,
        type: ArchiveStatusType = None,
        level: ArchiveStatusLevel = ArchiveStatusLevel.INFO,
        message: Optional[str] = None,
    ):
        """
        DEPRECATED: This method is deprecated and may be removed in future versions.
        Update the archiving or restore status of a dataset.

        :param dataset_id: The ICAT dataset identifier.
        :param type: The type of the status, possible options are {'archiving', 'restoration'}.
        :param level: The level of the status message; possible options are {'info', 'warning', 'error'}.
        :param message: The optional status' message.
        """
        raise NotImplementedError

    def update_metadata(
        self,
        proposal: str = None,
        beamline: str = None,
        dataset_paths: str = None,
        metadata_name: str = None,
        metadata_value: str = None,
    ):
        """
        Update or create datasets metadata.

        :param proposal: The proposal name.
        :param beamline: The beamline name of the proposal.
        :param dataset_paths: Comma-separated list of the dataset locations.
        :param metadata_name: The name of the metadata to update.
        :param metadata_value: The new value of the metadata.
        """
        raise NotImplementedError

    def add_files(
        self,
        dataset_id: int = None,
    ):
        """
        Add missing files to a dataset already ingested.

        :param dataset_id: The ICAT dataset identifier.
        """
        raise NotImplementedError

    def reschedule_investigation(self, investigation_id: str):
        """
        Reschedule an investigation defined by its id.

        :param investigation_id:
        """
        raise NotImplementedError

    def do_log_in(
        self, password: str, username: Optional[str] = None, plugin: str = "esrf"
    ) -> dict:
        """
        Login to access the restricted part of the API.

        :param password:
        :param username: optional username
        :param plugin: authentication plugin, defaults to 'esrf'
        :returns: authentication info
        """
        raise NotImplementedError

    def get_investigations_by(
        self,
        filter: Optional[str] = None,
        instrument_name: Optional[str] = None,
        start_date: Optional[datetime.datetime] = None,
        end_date: Optional[datetime.datetime] = None,
        ids: Optional[str] = None,
    ) -> List[dict]:
        """
        Returns a list of investigations matching the provided criteria.

        :param filter:
        :param instrument_name:
        :param start_date:
        :param end_date:
        :param ids:
        :returns: list of investigations
        """
        raise NotImplementedError

    def get_samples_by(
        self,
        investigation_id: Optional[str] = None,
        sample_ids: Optional[str] = None,
    ) -> List[dict]:
        """
        Returns a list of samples matching the provided criteria.

        :param investigation_id: an investigationId
        :param sample_ids: comma separated list of sampleIds
        """
        raise NotImplementedError

    def get_datasets_by(
        self,
        investigation_id: Optional[str] = None,
        dataset_ids: Optional[str] = None,
    ) -> List[dict]:
        """
        Returns a list of dataset matching the provided criteria.

        :param investigation_id: an investigationId
        :param dataset_ids: comma separated list of dataset Ids
        """
        raise NotImplementedError

    def get_parcels_by(self, investigation_id: str) -> List[dict]:
        """
        Returns the list of parcels associated to an investigation.

        :param investigation_id:
        :returns: list of parcel information
        """
        raise NotImplementedError

    def get_session_information(self) -> List[dict]:
        """
        Returns the information about a session.

        :returns: session information
        """
        raise NotImplementedError

    def get_sample_metadata_by(
        self,
        proposal: str,
        beamline: str,
        session_start_date: Optional[datetime.date] = None,
    ) -> List[SampleMetadata]:
        """
        Returns the list of sample metadata associated to a proposal and a beamline.

        :param proposal: Proposal identifier (e.g., "hg237").
        :param beamline: Beamline identifier (e.g., "id13").
        :param session_start_date: Experiment session start date (format yyyy-MM-dd).
        :returns: List of sample metadata.
        """
        raise NotImplementedError

    def get_sample_files_information_by(self, sample_id: str) -> dict:
        """
        Returns the sample file information for a given sample.

        :param sample_id: Sample identifier.
        :returns: Sample file information as dict.
        """
        raise NotImplementedError

    def download_file_by(
        self,
        sample_id: str,
        resource_id: str,
        use_chunks: bool = False,
        chunk_size: int = 8192,
    ) -> bytes:
        """
        Download a file associated to a sample.

        :param sample_id: Sample identifier.
        :param resource_id: Identifier of the resource/file to download.
        :param use_chunks: Read the file in chunks or the entire file into memory at once.
        :param chunk_size: Size of each chunk in bytes when `use_chunks` is True. Default is 8192 (8 KB).
        :returns: The content of the file as bytes.
        """
        raise NotImplementedError

    @property
    def expire_datasets_on_close(self) -> bool:
        """
        A flag indicating whether the dataset expires when it is closed or if it is synchronized with the metadata catalog.
        """
        raise NotImplementedError

    @property
    def reason_for_missing_information(self) -> str:
        """
        A string explaining why some information is missing in the metadata catalog.
        """
        raise NotImplementedError



import logging
import warnings
from typing import Type

logger = logging.getLogger(__name__)


class IcatMetadataFutureWarning(FutureWarning):
    """Warning for upcoming changes in ICAT metadata."""


class IcatMetadataSerializationError(TypeError):
    """ICAT metadata serialization failed."""


class IcatMetadataValidationError(ValueError):
    """ICAT metadata validation error when strict-mode is ON."""


class IcatMetadataValidationWarning(IcatMetadataFutureWarning):
    """Used instead of IcatMetadataValidationError when strict-mode is OFF."""


class IcatMetadataDeprecatedWarning(IcatMetadataFutureWarning):
    """Warning for deprecated ICAT metadata fields."""


def warn(message: str, warning_type: Type[Warning]) -> None:
    """Log message and emit as warning."""
    # For humans to see
    logger.warning(message)

    # For client code the act upon (e.g. log in the e-logbook)
    warnings.warn(message, warning_type, stacklevel=2)


def invalid_warn_or_raise(message: str, strict: bool = False) -> None:
    if strict:
        raise IcatMetadataValidationError(message)
    warn(message, IcatMetadataValidationWarning)



import os
from typing import List
from typing import Optional

from blissdata.beacon.files import read_config

from .main import IcatClient


def get_icat_client(
    timeout: Optional[float] = None, metadata_urls: Optional[List[str]] = None
) -> IcatClient:
    """
    Create and return an instance of `IcatClient` using configuration retrieved from a Beacon server.
    It is determined from the `BEACON_HOST` environment variable.
    If the configuration is unavailable, a `RuntimeError` is raised.
    :param timeout: optional timeout for the ICAT client operations.
    :param metadata_urls: URLs of the ActiveMQ message brokers to be used for creating
    ICAT datasets from a directory with metadata.
    If provided, these URLs will override the configuration retrieved from Beacon.
    :returns: the ICAT Client.
    """
    beacon_host = os.environ.setdefault("BEACON_HOST", "id00:25000")

    url = f"beacon://{beacon_host}/__init__.yml"
    config = read_config(url).get("icat_servers")

    if not config:
        raise RuntimeError(
            f"Beacon host {beacon_host} does not provide ICAT configuration"
        )
    return IcatClient(
        metadata_urls=metadata_urls if metadata_urls else config["metadata_urls"],
        elogbook_url=config["elogbook_url"],
        elogbook_token=config["elogbook_token"],
        feedback_timeout=timeout,
        add_files_urls=config["metadata_urls"],
    )



from datetime import date

import pytest
from requests.exceptions import HTTPError

from .utils import generate


def test_login(icatplus_restricted_client):
    client, messages = icatplus_restricted_client

    with pytest.raises(RuntimeError, match="Login is required."):
        _ = client.session_id

    with pytest.raises(HTTPError, match="403 Client Error: Authentication failed"):
        _ = client.login("wrong")

    result = client.login("correct")

    assert result["sessionId"] == client.session_id

    assert messages.empty()


def test_login_with_username_plugin(icatplus_restricted_client):
    client, messages = icatplus_restricted_client

    result = client.login(password="correct", username="john", plugin="db")

    assert result["sessionId"] == client.session_id
    assert messages.empty()


def test_get_investigations_by(icatplus_restricted_client, icat_metadata_client):
    client, messages = icatplus_restricted_client
    mclient, mmessages = icat_metadata_client

    with pytest.raises(RuntimeError, match="Login is required."):
        _ = client.get_investigations_by()

    assert client.login("correct")

    investigations = client.get_investigations_by()
    assert isinstance(investigations, list)

    mclient.start_investigation(proposal="hg123", beamline="id00")
    message = mmessages.get(timeout=10)

    investigations = client.get_investigations_by()
    expected = [
        {
            "experiment": "hg123",
            "id": 0,
            "instrument": {"name": "id00"},
            "proposal": "hg123",
            "startDate": message["startDate"],
        }
    ]
    assert investigations == expected

    assert messages.empty()
    assert mmessages.empty()


def test_get_parcels_by(icatplus_restricted_client):
    client, messages = icatplus_restricted_client

    investigation_id = generate.icat_id()

    with pytest.raises(RuntimeError, match="Login is required."):
        _ = client.get_parcels_by(investigation_id)

    assert client.login("correct")

    parcels = client.get_parcels_by(investigation_id)

    assert parcels == []
    assert messages.empty()


@pytest.mark.parametrize(
    "session_start_date, expected_ids",
    [
        (None, ["sa", "sb"]),
        (date(2020, 2, 1), ["sa"]),
    ],
)
def test_get_sample_metadata_by(
    icatplus_restricted_client, session_start_date, expected_ids
):
    client, messages = icatplus_restricted_client
    proposal = "hg237"
    beamline = "id13"

    with pytest.raises(RuntimeError, match="Login is required."):
        _ = client.get_sample_metadata_by(proposal, beamline, session_start_date)

    assert client.login("correct")

    samples = client.get_sample_metadata_by(proposal, beamline, session_start_date)

    returned_ids = [s["id"] for s in samples]
    assert returned_ids == expected_ids

    assert messages.empty()


def test_get_samples_by(icatplus_restricted_client):
    client, messages = icatplus_restricted_client

    investigation_id = generate.icat_id()

    with pytest.raises(RuntimeError, match="Login is required."):
        _ = client.get_samples_by(investigation_id)

    assert client.login("correct")

    samples = client.get_samples_by(investigation_id)

    assert samples == []
    assert messages.empty()


@pytest.mark.parametrize(
    "params, expected_ids",
    [
        ({"investigation_id": "mx415"}, ["sa", "sb"]),
        ({"investigation_id": "mx2012"}, ["sc"]),
        ({"sample_ids": "sc"}, ["sc"]),
        ({"investigation_id": "mx415", "sample_ids": "sa"}, ["sa"]),
        ({"investigation_id": "mx100"}, []),
        ({"sample_ids": "sd"}, []),
    ],
)
def test_get_samples_by_returns_expected_data(
    icatplus_restricted_client, params, expected_ids
):
    client, messages = icatplus_restricted_client
    client.login("correct")

    samples = client.get_samples_by(**params)
    returned_ids = [s["id"] for s in samples]

    assert returned_ids == expected_ids
    assert messages.empty()


def test_get_sample_files_information_by(icatplus_restricted_client):
    client, messages = icatplus_restricted_client

    sample_id = generate.icat_id()

    with pytest.raises(RuntimeError, match="Login is required."):
        _ = client.get_sample_files_information_by(sample_id)

    assert client.login("correct")

    sample_files = client.get_sample_files_information_by(sample_id)

    assert sample_files == []
    assert messages.empty()


@pytest.mark.parametrize(
    "use_chunks, chunk_size",
    [
        (False, None),
        (True, 1024),
    ],
)
def test_download_file_by_sample_parametrized(
    icatplus_restricted_client, use_chunks, chunk_size
):
    client, messages = icatplus_restricted_client

    sample_id = generate.icat_id()
    resource_id = generate.icat_id()

    if use_chunks:
        with pytest.raises(RuntimeError, match="Login is required."):
            _ = client.download_file_by(
                sample_id, resource_id, use_chunks=use_chunks, chunk_size=chunk_size
            )
    else:
        with pytest.raises(RuntimeError, match="Login is required."):
            _ = client.download_file_by(sample_id, resource_id)

    assert client.login("correct")

    result = client.download_file_by(
        sample_id, resource_id, use_chunks=use_chunks, chunk_size=chunk_size
    )

    assert isinstance(result, bytes)
    assert result == b"fake file content"
    assert messages.empty()



import concurrent.futures
import datetime
import json
import logging
import os
from glob import glob
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple

from tqdm import tqdm

from . import path_utils
from . import raw_data
from . import sync_types

logger = logging.getLogger("ICAT SYNC")


class ExperimentalSessionStore:
    def __init__(
        self,
        cache_dir: Optional[str] = None,
        save_dir: Optional[str] = None,
        raw_data_format: str = "esrfv3",
        invalidate_cache: bool = False,
    ) -> None:
        """
        :param cache_dir: one JSON file for each cached experimental session
        :param save_dir: one CSV+BASH file for each sync status
        :param raw_data_format: will be used as subdirectory name
        :param invalidate_cache: invalidating session when loaded from cache
        """
        self._save_dir = save_dir
        self._cache_dir = cache_dir
        self._raw_data_format = raw_data_format
        self._invalidate_cache = invalidate_cache

        self._save_subdir = None
        self._cache_subdir = None
        if save_dir is not None:
            self._save_subdir = os.path.join(save_dir, raw_data_format)
        if cache_dir is not None:
            self._cache_subdir = os.path.join(cache_dir, raw_data_format)

        self._exp_sessions: Dict[str, sync_types.ExperimentalSession] = dict()
        self._load_all_session_files()
        self._start_time = datetime.datetime.now()

    def __enter__(self) -> "ExperimentalSessionStore":
        self._start_time = datetime.datetime.now()
        return self

    def __exit__(self, *_) -> bool:
        self.close()
        return False

    def get_session(self, session_dir: str) -> Optional[sync_types.ExperimentalSession]:
        return self._exp_sessions.get(session_dir)

    def add_session(self, exp_session: sync_types.ExperimentalSession) -> None:
        """Add session to the in-memory cache and when enabled save it on disk."""
        self._exp_sessions[exp_session.session_dir] = exp_session
        if not self._cache_subdir:
            return
        filename = self._cache_filename(exp_session)
        data = exp_session.as_dict()
        os.makedirs(self._cache_subdir, mode=0o775, exist_ok=True)
        _json_save(data, filename)

    def remove_session(self, exp_session: sync_types.ExperimentalSession) -> None:
        """Remove session from the in-memory cache and when enabled from disk."""
        self._exp_sessions.pop(exp_session.session_dir, None)
        if not self._cache_subdir:
            return
        filename = self._cache_filename(exp_session)
        try:
            os.remove(filename)
        except FileNotFoundError:
            pass
        logger.info("Removed session cache %s", filename)

    def close(self) -> None:
        try:
            self._save_overview()
        finally:
            self._print_final()

    def _load_all_session_files(self) -> None:
        """Load all the session files when enabled and delete the ones for sessions which directory no longer exists."""
        if not self._cache_subdir:
            return
        logger.info("Loading cache from %s ...", self._cache_subdir)
        filenames = glob(os.path.join(self._cache_subdir, "*.json"))

        logger.info("Loading cache of %d sessions ...", len(filenames))
        data = _executor_map_with_progress(_json_load, filenames)

        if self._invalidate_cache:
            logger.info("Invalidate cache of %d sessions ...", len(data))

        if self._invalidate_cache:
            func = _parse_session_cache_with_invalidation
        else:
            func = _parse_session_cache
        exp_sessions = _executor_map_with_progress(func, filenames, data)

        for exp_session in exp_sessions:
            if exp_session is not None:
                self._exp_sessions[exp_session.session_dir] = exp_session

        logger.info("Loaded cache of %d sessions.", len(self._exp_sessions))

    def _cache_filename(self, exp_session: sync_types.ExperimentalSession) -> str:
        basename = (
            f"{exp_session.proposal}_{exp_session.beamline}_{exp_session.session}.json"
        )
        return os.path.join(self._cache_subdir, basename)

    def _save_overview(self) -> None:
        if not self._save_subdir:
            return
        logger.info(
            "Saving synchronization overview of %d sessions ...",
            len(self._exp_sessions),
        )
        os.makedirs(self._save_subdir, mode=0o775, exist_ok=True)

        save_columns = [
            "beamline",
            "proposal",
            "timeslot",
            "session",
            "url",
            "search_url",
            "path",
            *sync_types.ExperimentalSession.DATASET_STATUSES,
        ]

        save_category = [
            "todo",
            "ok",
            "empty",
            "has_invalid",
            "todo_ongoing",
            "not_in_icat",
        ]

        register_cmd_params = ""
        if self._save_dir:
            register_cmd_params += f"--save-dir {self._save_dir} "
        if self._cache_dir:
            register_cmd_params += f"--cache-dir {self._cache_dir} "
        register_cmd_params += '"$@"'

        with concurrent.futures.ThreadPoolExecutor() as executor:
            # Clear CSV files and add header
            stats_files = dict(
                (s, os.path.join(self._save_subdir, s + ".csv")) for s in save_category
            )

            def _init_csv(filename):
                _save_csv_line(filename, save_columns, clear=True)
                try:
                    os.chmod(filename, 0o0664)
                except PermissionError:
                    pass

            _ = list(executor.map(_init_csv, stats_files.values()))

            # Clear bash files and add header
            cmd_files = dict(
                (s, os.path.join(self._save_subdir, s + ".sh")) for s in save_category
            )

            def _init_cmd(filename):
                _save_file_line(filename, "#!/bin/bash", clear=True)
                _save_file_line(filename, "")
                _save_file_line(
                    filename,
                    "# Run a cached copy because this script can be modified during execution",
                )
                _save_file_line(filename, 'if [ -z "$SCRIPT_CACHED" ]; then')
                _save_file_line(filename, '    SCRIPT_CACHED=$(<"$0")')
                _save_file_line(filename, '    script_name=$(basename "$0")')
                _save_file_line(
                    filename, '    echo "$SCRIPT_CACHED" > /tmp/cached_$script_name'
                )
                _save_file_line(
                    filename,
                    '    SCRIPT_CACHED=1 /bin/bash /tmp/cached_$script_name "$@"',
                )
                _save_file_line(filename, "    exit $?")
                _save_file_line(filename, "fi")
                _save_file_line(filename, "")
                try:
                    os.chmod(filename, 0o0774)
                except PermissionError:
                    pass

            _ = list(executor.map(_init_cmd, cmd_files.values()))

            # Add sessions CSV and BASH files
            def _save_session_line(exp_session):
                icat_investigation = exp_session.icat_investigation
                if not icat_investigation:
                    save_category = "not_in_icat"
                else:
                    statuses = exp_session.dataset_statuses()
                    if statuses & {
                        "not_uploaded",
                        "unregistered",
                        "registered_without_files",
                        "excluded_too_many_files",
                    }:
                        if icat_investigation and icat_investigation.ongoing:
                            save_category = "todo_ongoing"
                        else:
                            save_category = "todo"
                    elif "invalid" in statuses:
                        save_category = "has_invalid"
                    elif "registered" in statuses:
                        save_category = "ok"
                    else:
                        save_category = "empty"

                tabular_data = exp_session.tabular_data()
                stats = [tabular_data[k] for k in save_columns]
                _save_csv_line(stats_files[save_category], stats)

                register_cmd = f"icat-sync-raw --beamline {exp_session.beamline} --proposal {exp_session.proposal} --session {exp_session.session} {register_cmd_params}"
                _save_file_line(cmd_files[save_category], register_cmd)

            _ = list(executor.map(_save_session_line, self._exp_sessions.values()))

        logger.info("Saving synchronization finished.")

    def _print_final(self):
        end_time = datetime.datetime.now()
        logger.info("")
        logger.info("Total synchronization time was %s", end_time - self._start_time)
        if self._cache_subdir:
            logger.info("Cache was used from %s", self._cache_subdir)
        if self._save_subdir:
            logger.info("Results can be found in %s", self._save_subdir)
            logger.info(
                "To fix issues run %s", os.path.join(self._save_subdir, "todo.sh")
            )


def _executor_map_with_progress(func: Callable, *args) -> List[Any]:
    results = []
    with concurrent.futures.ThreadPoolExecutor() as executor:
        with tqdm(total=len(args[0])) as progress:

            def update_progress(future):
                progress.update(1)
                results.append(future.result())

            futures = [executor.submit(func, *_args) for _args in zip(*args)]

            for future in futures:
                future.add_done_callback(update_progress)

            result = concurrent.futures.wait(futures)

            for future in result.done:
                if future.exception():
                    raise future.exception()

    return results


def _parse_session_cache(
    filename: str, data: dict
) -> Optional[sync_types.ExperimentalSession]:
    try:
        return sync_types.ExperimentalSession.from_dict(data)
    except Exception as e:
        logger.info("Invalidate %s: wrong cache format (%s)", filename, e)
        try:
            os.remove(filename)
        except FileNotFoundError:
            pass


def _parse_session_cache_with_invalidation(
    filename: str, data: dict
) -> Optional[sync_types.ExperimentalSession]:
    exp_session = _parse_session_cache(filename, data)
    if exp_session is None:
        return

    invalidation_reason = _session_invalidation_reason(exp_session)
    if invalidation_reason:
        logger.info("Invalidate %s: %s", exp_session.session_dir, invalidation_reason)
        try:
            os.remove(filename)
        except FileNotFoundError:
            pass
        return

    return exp_session


def _session_invalidation_reason(
    exp_session: sync_types.ExperimentalSession,
) -> Optional[str]:
    if not os.path.isdir(exp_session.session_dir):
        return "no longer exists"

    if not exp_session.icat_investigation:
        # Cached session does not contain any datasets when
        # no ICAT investigation was found
        return

    cached = {path_utils.markdir(dset.path) for dset in exp_session.iter_datasets()}
    dataset_filters = raw_data.get_dataset_filters(
        exp_session.raw_root_dir, raw_data_format=exp_session.raw_data_format
    )
    actual = {
        path_utils.markdir(path)
        for dataset_filter in dataset_filters
        for path in glob(path_utils.markdir(dataset_filter))
    }
    if cached != actual:
        nnew = len(actual - cached)
        nremoved = len(cached - actual)
        return f"has {nnew} new datasets and {nremoved} datasets were removed"


def _json_save(data: Dict[str, Any], filename: str) -> None:
    try:
        with open(filename, "w") as fp:
            json.dump(data, fp, default=_json_serializer, indent=2)
    except Exception as e:
        raise IOError(f"writing failed for file {filename}") from e


def _json_load(filename: str) -> Dict[str, Any]:
    try:
        with open(filename, "r") as fp:
            return json.load(fp, object_pairs_hook=_json_pair_deserializer)
    except Exception as e:
        raise IOError(f"reading failed for file {filename}") from e


def _json_serializer(obj: Any):
    """Serialize datetime and date objects to string."""
    if isinstance(obj, (datetime.datetime, datetime.date)):
        return obj.isoformat()
    else:
        raise TypeError(
            "Object of type '{}' is not JSON serializable".format(type(obj))
        )


def _json_pair_deserializer(items: List[Tuple[str, Any]]) -> Dict[str, Any]:
    """Deserialize string to datetime or date objects."""
    result = dict()
    for key, value in items:
        if "date" in key and value:
            try:
                value = datetime.date.fromisoformat(value)
            except ValueError:
                value = datetime.datetime.fromisoformat(value)
        result[key] = value
    return result


def _save_csv_line(filename: str, row: List[Any], clear: bool = False) -> None:
    line = ",".join([str(item).replace(",", " ").strip() for item in row])
    _save_file_line(filename, line, clear)


def _save_file_line(filename: str, line: str, clear: bool = False) -> None:
    mode = "w" if clear else "a"
    with open(filename, mode) as f:
        f.write(f"{line}\n")



import http.server
import json
import logging
import os
import re
import socket
import urllib.parse
from typing import Iterator
from typing import Optional
from typing import Tuple

from ...utils.log_utils import basic_config
from .icat_db import IcatDb
from .icat_session import IcatSession
from .utils import ReuseAddrTCPServer

logger = logging.getLogger("ICATPLUS SERVER")


class MyTCPRequestHandler(http.server.BaseHTTPRequestHandler):
    def __init__(self, *args, s_out=None, icat_data_dir: Optional[str] = None, **kw):
        self.s_out = s_out
        self.icatdb = IcatDb(icat_data_dir)
        self.icatsession = IcatSession(icat_data_dir)
        self._init_fake_data()
        super().__init__(*args, **kw)

    def _init_fake_data(self):
        self.fake_sample_data = {
            "mx415": [
                {"id": "sa", "name": "Sample A"},
                {"id": "sb", "name": "Sample B"},
            ],
            "mx2012": [
                {"id": "sc", "name": "Sample C"},
            ],
        }
        self.fake_samplemetadata_data = {
            "all_sessions": [
                {
                    "id": "sa",
                    "name": "Sample A",
                    "linkedSampleAFormName": None,
                    "proposal": "hg237",
                    "beamline": "id13",
                    "sessionStartDate": "2020-02-01",
                },
                {
                    "id": "sb",
                    "name": "Sample B",
                    "linkedSampleAFormName": None,
                    "proposal": "hg237",
                    "beamline": "id13",
                    "sessionStartDate": "2020-03-01",
                },
            ],
            "2020-02-01": [
                {
                    "id": "sa",
                    "name": "Sample A",
                    "linkedSampleAFormName": None,
                    "proposal": "hg237",
                    "beamline": "id13",
                    "sessionStartDate": "2020-02-01",
                },
            ],
        }

    def do_HEAD(self):  # noqa N802
        logger.info("HEAD")
        self.reply_ok()

    def do_GET(self):  # noqa N802
        if self._handle_investigation():
            return
        if self._handle_dataset():
            return
        if self._handle_get_investigations_by():
            return
        if self._handle_get_parcels_by():
            return
        if self._handle_get_sample_metadata_by():
            return
        if self._handle_get_samples_by():
            return
        if self._handle_get_sample_files_information_by():
            return
        if self._handle_download_file_by():
            return
        logger.error("Bad GET request: %s", self.path)
        self.reply_bad_request()

    def do_POST(self):  # noqa N802
        if self._handle_elogbook_message():
            return
        if self._handle_elogbook_data():
            return
        if self._handle_elogbook_message_beamline_only():
            return
        if self._handle_elogbook_message_no_beamline():
            return
        if self._handle_elogbook_data_beamline_only():
            return
        if self._handle_login():
            return
        logger.error("Bad POST request: %s", self.path)
        self.reply_bad_request()

    def _handle_investigation(self):
        fmt = "/dataacquisition/(?P<apikey>[^//]+)/investigation\\?instrumentName=(?P<instrument>[^//]+)&investigationName=(?P<investigation>[^//]+)&?(?P<parameters>[^//]+)?"
        mresult = re.match(fmt, self.path)
        if not mresult:
            return
        data = mresult.groupdict()
        if data["parameters"]:
            data["parameters"] = dict(
                s.split("=") for s in data["parameters"].split("&")
            )

        if "666" in data["investigation"]:
            self.reply_error()
            return True

        investigations = self.icatdb.get_investigations(
            data["instrument"], data["investigation"]
        )
        # TODO: use parameters to filter
        logger.info("Investigation GET response: %s", investigations)

        self.reply_json(investigations)
        return True

    def _iter_investigations(self) -> Iterator[Tuple[int, dict]]:
        root_dir = os.path.join(self.icat_data_dir, "investigations")
        if not os.path.isdir(root_dir):
            return
        for basename in os.listdir(root_dir):
            filename = os.path.join(root_dir, basename)
            investigation_id = int(os.path.splitext(basename)[0])
            with open(filename, "r") as f:
                investigation = json.load(f)
            yield investigation_id, investigation

    def _handle_dataset(self):
        fmt = "/dataacquisition/(?P<apikey>[^//]+)/dataset\\?investigationId=(?P<investigation_id>[^//]+)&?(?P<parameters>[^//]+)?"
        mresult = re.match(fmt, self.path)
        if not mresult:
            return
        data = mresult.groupdict()
        if data["parameters"]:
            data["parameters"] = dict(
                s.split("=") for s in data["parameters"].split("&")
            )

        datasets = self.icatdb.get_datasets(mresult["investigation_id"])
        # TODO: use parameters to filter
        logger.info("Dataset GET response: %s", datasets)
        self.reply_json(datasets)
        return True

    def _handle_elogbook_message(self):
        logger.info(self.path)
        fmt = "/dataacquisition/(?P<apikey>[^//]+)/notification\\?instrumentName=(?P<instrument>[^//]+)&investigationName=(?P<investigation>[^//]+)"
        mresult = re.match(fmt, self.path)
        if not mresult:
            return
        if (
            self.headers.get("content-type") != "application/json"
            or "666" in mresult.groupdict()["investigation"]
        ):
            self.reply_bad_request()
            return True
        length = int(self.headers.get("content-length"))
        adict = json.loads(self.rfile.read(length))
        adict.update(mresult.groupdict())
        self.on_elog_message(adict)
        self.reply_ok()
        return True

    def _handle_elogbook_data(self):
        fmt = "/dataacquisition/(?P<apikey>[^//]+)/base64\\?instrumentName=(?P<instrument>[^//]+)&investigationName=(?P<investigation>[^//]+)"
        mresult = re.match(fmt, self.path)
        if not mresult:
            return
        if (
            self.headers.get("content-type") != "application/json"
            or "666" in mresult.groupdict()["investigation"]
        ):
            self.reply_bad_request()
            return True
        length = int(self.headers.get("content-length"))
        adict = json.loads(self.rfile.read(length))
        adict.update(mresult.groupdict())
        self.on_elog_message(adict)
        self.reply_ok()
        return True

    def _handle_elogbook_message_beamline_only(self):
        logger.info(self.path)
        fmt = "/dataacquisition/(?P<apikey>[^//]+)/notification\\?instrumentName=(?P<instrument>[^//]+)"
        mresult = re.match(fmt, self.path)
        if not mresult:
            return
        if self.headers.get("content-type") != "application/json":
            self.reply_bad_request()
            return True
        length = int(self.headers.get("content-length"))
        adict = json.loads(self.rfile.read(length))
        adict.update(mresult.groupdict())
        self.on_elog_message(adict)
        self.reply_ok()
        return True

    def _handle_elogbook_message_no_beamline(self):
        logger.info(self.path)
        fmt = "/dataacquisition/(?P<apikey>[^//]+)/notification"
        mresult = re.match(fmt, self.path)
        if not mresult:
            return
        if self.headers.get("content-type") != "application/json":
            self.reply_bad_request()
            return True
        length = int(self.headers.get("content-length"))
        adict = json.loads(self.rfile.read(length))
        adict.update(mresult.groupdict())
        self.on_elog_message(adict)
        self.reply_ok()
        return True

    def _handle_elogbook_data_beamline_only(self):
        fmt = "/dataacquisition/(?P<apikey>[^//]+)/base64\\?instrumentName=(?P<instrument>[^//]+)"
        mresult = re.match(fmt, self.path)
        if not mresult:
            return
        if self.headers.get("content-type") != "application/json":
            self.reply_bad_request()
            return True
        length = int(self.headers.get("content-length"))
        adict = json.loads(self.rfile.read(length))
        adict.update(mresult.groupdict())
        self.on_elog_message(adict)
        self.reply_ok()
        return True

    def _handle_login(self):
        fmt = "/session"
        mresult = re.match(fmt, self.path)
        if not mresult:
            return
        if self.headers.get("content-type") != "application/json":
            self.reply_bad_request()
            return True
        length = int(self.headers.get("content-length"))
        credentials = json.loads(self.rfile.read(length))

        try:
            icat_session = self.icatsession.login(credentials)
        except PermissionError:
            self.reply_forbidden(message="Authentication failed")
        else:
            self.reply_json(icat_session)

        return True

    def _handle_get_investigations_by(self):
        fmt = r"catalogue/([a-f0-9-]{36})/investigation"
        mresult = re.search(fmt, self.path)
        if not mresult:
            return
        session_id = mresult.group(1)
        if self.icatsession.is_allowed(session_id):
            self.reply_json(self.icatdb.get_all_investigations())
        else:
            self.reply_forbidden(message="Authentication failed")
        return True

    def _handle_get_parcels_by(self):
        fmt = r"tracking/([a-f0-9-]{36})/parcel\?investigationId=(\d+)"
        mresult = re.search(fmt, self.path)
        if not mresult:
            return
        session_id = mresult.group(1)
        if self.icatsession.is_allowed(session_id):
            self.reply_json(list())
        else:
            self.reply_forbidden(message="Authentication failed")
        return True

    def _handle_get_sample_metadata_by(self):
        fmt = r"samplemetadata/([a-f0-9-]{36})/samples/acquisition(?:\?(.*))?"
        mresult = re.search(fmt, self.path)
        if not mresult:
            return
        session_id, query = mresult.groups()

        if not self.icatsession.is_allowed(session_id):
            self.reply_forbidden(message="Authentication failed")
            return True
        params = dict(urllib.parse.parse_qsl(query or ""))
        sample_metadata = []
        start_date = params.get("startDate")

        if start_date is None:
            sample_metadata = self.fake_samplemetadata_data.get("all_sessions", [])
        else:
            sample_metadata = self.fake_samplemetadata_data.get(start_date, [])

        self.reply_json(sample_metadata)
        return True

    def _handle_get_samples_by(self):
        fmt = r"catalogue/([a-f0-9-]{36})/samples(?:\?(.*))?"
        mresult = re.search(fmt, self.path)
        if not mresult:
            return
        session_id, query = mresult.groups()

        if not self.icatsession.is_allowed(session_id):
            self.reply_forbidden(message="Authentication failed")
            return True

        params = dict(urllib.parse.parse_qsl(query or ""))

        samples = []
        requested_sample_ids = None
        if "sampleIds" in params:
            requested_sample_ids = [s.strip() for s in params["sampleIds"].split(",")]
        if "investigationId" in params:
            investigation_id = params["investigationId"]
            samples = self.fake_sample_data.get(investigation_id, [])
            if requested_sample_ids:
                samples = [s for s in samples if s["id"] in requested_sample_ids]
        elif "sampleIds" in params:
            all_samples = [s for group in self.fake_sample_data.values() for s in group]
            samples = [s for s in all_samples if s["id"] in requested_sample_ids]
        self.reply_json(samples)
        return True

    def _handle_get_sample_files_information_by(self):
        fmt = r"catalogue/([a-f0-9-]{36})/files\?sampleId=(\d+)"
        mresult = re.search(fmt, self.path)
        if not mresult:
            return
        session_id = mresult.group(1)
        if self.icatsession.is_allowed(session_id):
            self.reply_json(list())
        else:
            self.reply_forbidden(message="Authentication failed")
        return True

    def _handle_download_file_by(self):
        fmt = (
            r"catalogue/([a-f0-9-]{36})/files/download\?sampleId=(\d+)&resourceId=(\d+)"
        )
        mresult = re.search(fmt, self.path)
        if not mresult:
            return
        session_id = mresult.group(1)
        if not self.icatsession.is_allowed(session_id):
            self.reply_forbidden(message="Authentication failed")
            return True
        fake_content = b"fake file content"
        filename = "testfile.txt"
        self.send_response(200)
        self.send_header("Content-Type", "application/octet-stream")
        self.send_header("Content-Disposition", f'attachment; filename="{filename}"')
        self.send_header("Content-Length", str(len(fake_content)))
        self.end_headers()
        self.wfile.write(fake_content)
        return True

    def reply_ok(self):
        self.send_response(http.HTTPStatus.OK)
        self.end_headers()

    def reply_bad_request(self):
        self.send_response(http.HTTPStatus.BAD_REQUEST)
        self.end_headers()

    def reply_not_found(self):
        self.send_response(http.HTTPStatus.NOT_FOUND)
        self.end_headers()

    def reply_error(self):
        self.send_response(http.HTTPStatus.INTERNAL_SERVER_ERROR)
        self.end_headers()

    def reply_forbidden(self, **kw):
        self.send_response(http.HTTPStatus.FORBIDDEN, **kw)
        self.end_headers()

    def reply_json(self, adict):
        body = json.dumps(adict).encode()
        self.send_response(http.HTTPStatus.OK)
        self.send_header("Content-type", "application/json")
        self.send_header("Content-Length", str(len(body)))
        self.end_headers()
        self.wfile.write(body)

    def on_elog_message(self, adict):
        if not adict:
            return
        logger.info("received message:\n %s", adict)
        if self.s_out is not None:
            self.s_out.sendall(json.dumps(adict).encode() + b"\n")


def main(port=8443, port_out=0, icat_data_dir: Optional[str] = None):
    if port_out:
        s_out = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s_out.connect(("localhost", port_out))
    else:
        s_out = None

    class MyTCPRequestHandlerWithSocket(MyTCPRequestHandler):
        def __init__(self, *args, **kw):
            super().__init__(*args, s_out=s_out, icat_data_dir=icat_data_dir, **kw)

    # Create a TCP Server instance
    server_instance = ReuseAddrTCPServer(
        ("localhost", port), MyTCPRequestHandlerWithSocket
    )

    # Start accepting requests and setup output socket
    if port_out:
        logger.info(f"Redirect received messages to port {port_out}")
        s_out.sendall(json.dumps({"STATUS": "LISTENING"}).encode() + b"\n")
    try:
        logger.info("CTRL-C to stop")
        server_instance.serve_forever()
    finally:
        if port_out:
            s_out.close()
        logger.info("Exit.")


if __name__ == "__main__":
    import argparse

    basic_config(
        logger=logger,
        level=logging.DEBUG,
        format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
    )

    parser = argparse.ArgumentParser(description="ICAT plus server")
    parser.add_argument("--port", default=8443, type=int, help="server port")
    parser.add_argument("--port_out", default=0, type=int, help="output socket")
    parser.add_argument(
        "--icat_data_dir", default=None, type=str, help="One file per dataset"
    )
    args = parser.parse_args()
    main(port=args.port, port_out=args.port_out, icat_data_dir=args.icat_data_dir)



import datetime
import logging
import os
import re

import pytest

from ... import errors
from ..._icat_messaging.serializers.dataset import raw_serialize_dataset_message
from ..._icat_messaging.serializers.dataset import serialize_dataset_message
from ..._icat_messaging.serializers.dataset_parameters.main import (
    serialize_dataset_parameters,
)
from ..._icat_messaging.serializers.investigation import (
    raw_serialize_investigation_message,
)
from ..._icat_messaging.utils import datetime_utils


def test_dataset_message():
    proposal = "MX-123"
    beamline = "ID30A-3"
    start_datetime = datetime.datetime.now().astimezone()
    end_datetime = start_datetime + datetime.timedelta(minutes=1)
    path = "/data/run1/"
    dataset = "Protein Crystal Diffraction Run"
    metadata = {
        "machine": "mymachine",
        "software": "mysoftware_version",
        "startDate": start_datetime,
        "endDate": end_datetime,
        "Sample_name": "Lysozyme Crystal",
    }

    expected = f"""<?xml version="1.0" encoding="UTF-8"?>
<tns:dataset xmlns:tns="http://www.esrf.fr/icat" complete="true">
  <tns:investigation>MX-123</tns:investigation>
  <tns:instrument>ID30A-3</tns:instrument>
  <tns:name>Protein Crystal Diffraction Run</tns:name>
  <tns:location>{os.path.normpath(path)}</tns:location>
  <tns:startDate>{datetime_utils.with_timezone_zulu_format(start_datetime)}</tns:startDate>
  <tns:endDate>{datetime_utils.with_timezone_zulu_format(end_datetime)}</tns:endDate>
  <tns:sample>
    <tns:name>Lysozyme Crystal</tns:name>
  </tns:sample>
  <tns:parameter>
    <tns:name>Sample_name</tns:name>
    <tns:value>Lysozyme Crystal</tns:value>
  </tns:parameter>
  <tns:parameter>
    <tns:name>datasetName</tns:name>
    <tns:value>Protein Crystal Diffraction Run</tns:value>
  </tns:parameter>
  <tns:parameter>
    <tns:name>proposal</tns:name>
    <tns:value>MX-123</tns:value>
  </tns:parameter>
  <tns:parameter>
    <tns:name>location</tns:name>
    <tns:value>{os.path.normpath(path)}</tns:value>
  </tns:parameter>
  <tns:parameter>
    <tns:name>startDate</tns:name>
    <tns:value>{datetime_utils.with_timezone_zulu_format(start_datetime)}</tns:value>
  </tns:parameter>
  <tns:parameter>
    <tns:name>endDate</tns:name>
    <tns:value>{datetime_utils.with_timezone_zulu_format(end_datetime)}</tns:value>
  </tns:parameter>
  <tns:parameter>
    <tns:name>machine</tns:name>
    <tns:value>mymachine</tns:value>
  </tns:parameter>
  <tns:parameter>
    <tns:name>software</tns:name>
    <tns:value>mysoftware_version</tns:value>
  </tns:parameter>
  <tns:parameter>
    <tns:name>complete</tns:name>
    <tns:value>true</tns:value>
  </tns:parameter>
</tns:dataset>
"""

    data = serialize_dataset_message(
        beamline=beamline,
        proposal=proposal,
        dataset=dataset,
        path=path,
        metadata=metadata,
        indent="  ",
    )
    assert data.decode("utf-8") == expected


def test_dataset_message_unknown_parameter():
    proposal = "MX-123"
    beamline = "ID30A-3"
    start_datetime = datetime.datetime.now().astimezone()
    end_datetime = start_datetime + datetime.timedelta(minutes=1)
    path = "/data/run1"
    dataset = "Protein Crystal Diffraction Run"
    metadata = {
        "machine": "mymachine",
        "software": "mysoftware_version",
        "startDate": start_datetime,
        "endDate": end_datetime,
        "Sample_name": "Lysozyme Crystal",
        "InstrumentBeam_incident_energy": (7000, "eV"),  # partial validation
        "Unknown_key": 42,  # fails
    }

    expected = f"""<?xml version="1.0" encoding="UTF-8"?>
<tns:dataset xmlns:tns="http://www.esrf.fr/icat" complete="true">
  <tns:investigation>MX-123</tns:investigation>
  <tns:instrument>ID30A-3</tns:instrument>
  <tns:name>Protein Crystal Diffraction Run</tns:name>
  <tns:location>{os.path.normpath(path)}</tns:location>
  <tns:startDate>{datetime_utils.with_timezone_zulu_format(start_datetime)}</tns:startDate>
  <tns:endDate>{datetime_utils.with_timezone_zulu_format(end_datetime)}</tns:endDate>
  <tns:sample>
    <tns:name>Lysozyme Crystal</tns:name>
  </tns:sample>
  <tns:parameter>
    <tns:name>Sample_name</tns:name>
    <tns:value>Lysozyme Crystal</tns:value>
  </tns:parameter>
  <tns:parameter>
    <tns:name>InstrumentBeam_incident_energy</tns:name>
    <tns:value>7.0</tns:value>
  </tns:parameter>
  <tns:parameter>
    <tns:name>datasetName</tns:name>
    <tns:value>Protein Crystal Diffraction Run</tns:value>
  </tns:parameter>
  <tns:parameter>
    <tns:name>proposal</tns:name>
    <tns:value>MX-123</tns:value>
  </tns:parameter>
  <tns:parameter>
    <tns:name>location</tns:name>
    <tns:value>{os.path.normpath(path)}</tns:value>
  </tns:parameter>
  <tns:parameter>
    <tns:name>startDate</tns:name>
    <tns:value>{datetime_utils.with_timezone_zulu_format(start_datetime)}</tns:value>
  </tns:parameter>
  <tns:parameter>
    <tns:name>endDate</tns:name>
    <tns:value>{datetime_utils.with_timezone_zulu_format(end_datetime)}</tns:value>
  </tns:parameter>
  <tns:parameter>
    <tns:name>machine</tns:name>
    <tns:value>mymachine</tns:value>
  </tns:parameter>
  <tns:parameter>
    <tns:name>software</tns:name>
    <tns:value>mysoftware_version</tns:value>
  </tns:parameter>
  <tns:parameter>
    <tns:name>complete</tns:name>
    <tns:value>true</tns:value>
  </tns:parameter>
  <tns:parameter>
    <tns:name>Unknown_key</tns:name>
    <tns:value>42</tns:value>
  </tns:parameter>
</tns:dataset>
"""

    expected_message = _prepare_regex_pattern(r"""Unknown ICAT fields: ['Unknown_key']

Documentation on ICAT fields: __ICAT_DEF_URL__""")

    with pytest.warns(errors.IcatMetadataValidationWarning, match=expected_message):
        data = serialize_dataset_message(
            beamline=beamline,
            proposal=proposal,
            dataset=dataset,
            path=path,
            metadata=metadata,
            indent="  ",
        )
    assert data.decode("utf-8") == expected

    with pytest.raises(errors.IcatMetadataValidationError, match=expected_message):
        _ = serialize_dataset_message(
            beamline=beamline,
            proposal=proposal,
            dataset=dataset,
            path=path,
            metadata=metadata,
            indent="  ",
            strict=True,
        )


def test_dataset_message_wrong_value():
    proposal = "MX-123"
    beamline = "ID30A-3"
    start_datetime = datetime.datetime.now().astimezone()
    end_datetime = start_datetime + datetime.timedelta(minutes=1)
    path = "/data/run1"
    dataset = "Protein Crystal Diffraction Run"
    metadata = {
        "machine": "mymachine",
        "software": "mysoftware_version",
        "startDate": start_datetime,
        "endDate": end_datetime,
        "Sample_name": "Lysozyme Crystal",
        "InstrumentBeam_incident_energy": (7000, "eV"),  # partial validation
        "InstrumentBeam_distance": "wrong",  # fails
    }

    expected = f"""<?xml version="1.0" encoding="UTF-8"?>
<tns:dataset xmlns:tns="http://www.esrf.fr/icat" complete="true">
  <tns:investigation>MX-123</tns:investigation>
  <tns:instrument>ID30A-3</tns:instrument>
  <tns:name>Protein Crystal Diffraction Run</tns:name>
  <tns:location>{os.path.normpath(path)}</tns:location>
  <tns:startDate>{datetime_utils.with_timezone_zulu_format(start_datetime)}</tns:startDate>
  <tns:endDate>{datetime_utils.with_timezone_zulu_format(end_datetime)}</tns:endDate>
  <tns:sample>
    <tns:name>Lysozyme Crystal</tns:name>
  </tns:sample>
  <tns:parameter>
    <tns:name>Sample_name</tns:name>
    <tns:value>Lysozyme Crystal</tns:value>
  </tns:parameter>
  <tns:parameter>
    <tns:name>InstrumentBeam_incident_energy</tns:name>
    <tns:value>7.0</tns:value>
  </tns:parameter>
  <tns:parameter>
    <tns:name>datasetName</tns:name>
    <tns:value>Protein Crystal Diffraction Run</tns:value>
  </tns:parameter>
  <tns:parameter>
    <tns:name>proposal</tns:name>
    <tns:value>MX-123</tns:value>
  </tns:parameter>
  <tns:parameter>
    <tns:name>location</tns:name>
    <tns:value>{os.path.normpath(path)}</tns:value>
  </tns:parameter>
  <tns:parameter>
    <tns:name>startDate</tns:name>
    <tns:value>{datetime_utils.with_timezone_zulu_format(start_datetime)}</tns:value>
  </tns:parameter>
  <tns:parameter>
    <tns:name>endDate</tns:name>
    <tns:value>{datetime_utils.with_timezone_zulu_format(end_datetime)}</tns:value>
  </tns:parameter>
  <tns:parameter>
    <tns:name>machine</tns:name>
    <tns:value>mymachine</tns:value>
  </tns:parameter>
  <tns:parameter>
    <tns:name>software</tns:name>
    <tns:value>mysoftware_version</tns:value>
  </tns:parameter>
  <tns:parameter>
    <tns:name>complete</tns:name>
    <tns:value>true</tns:value>
  </tns:parameter>
  <tns:parameter>
    <tns:name>InstrumentBeam_distance</tns:name>
    <tns:value>wrong</tns:value>
  </tns:parameter>
</tns:dataset>
"""

    expected_message = _prepare_regex_pattern(
        r"""ValidationError: 1 validation error for IcatDatasetParameters
instrument.beam.distance
  Value error, could not convert string to float: 'wrong' [type=value_error, input_value='wrong', input_type=str]
    For further information visit __VALUE_ERROR_URL__

Documentation on ICAT fields: __ICAT_DEF_URL__"""
    )

    with pytest.warns(errors.IcatMetadataValidationWarning, match=expected_message):
        data = serialize_dataset_message(
            beamline=beamline,
            proposal=proposal,
            dataset=dataset,
            path=path,
            metadata=metadata,
            indent="  ",
        )
    assert data.decode("utf-8") == expected

    with pytest.raises(errors.IcatMetadataValidationError):
        _ = serialize_dataset_message(
            beamline=beamline,
            proposal=proposal,
            dataset=dataset,
            path=path,
            metadata=metadata,
            indent="  ",
            strict=True,
        )


def test_dataset_message_xsd_only():
    data = {
        "datasetId": 1001,
        "investigation": "MX-123",
        "instrument": "ID30A-3",
        "name": "Protein Crystal Diffraction Run",
        "location": "/data/run1",
        "startDate": "2026-03-08T09:18:49.678825+01:00",
        "endDate": "2026-03-08T09:19:49.678825+01:00",
        "sample": {
            "name": "Lysozyme Crystal",
            "type": "Protein Crystal",
            "parameter": [
                {"name": "temperature", "value": "100K"},
                {"name": "buffer", "value": "NaCl 0.1M"},
            ],
        },
        "datafile": [
            {
                "location": "/data/run1/file1.cbf",
                "size": 12345,
            }
        ],
    }

    expected = """<?xml version="1.0" encoding="UTF-8"?>
<tns:dataset xmlns:tns="http://www.esrf.fr/icat" complete="true">
  <tns:datasetId>1001</tns:datasetId>
  <tns:investigation>MX-123</tns:investigation>
  <tns:instrument>ID30A-3</tns:instrument>
  <tns:name>Protein Crystal Diffraction Run</tns:name>
  <tns:location>/data/run1</tns:location>
  <tns:startDate>2026-03-08T09:18:49.678825+01:00</tns:startDate>
  <tns:endDate>2026-03-08T09:19:49.678825+01:00</tns:endDate>
  <tns:sample>
    <tns:name>Lysozyme Crystal</tns:name>
    <tns:type>Protein Crystal</tns:type>
    <tns:parameter>
      <tns:name>temperature</tns:name>
      <tns:value>100K</tns:value>
    </tns:parameter>
    <tns:parameter>
      <tns:name>buffer</tns:name>
      <tns:value>NaCl 0.1M</tns:value>
    </tns:parameter>
  </tns:sample>
  <tns:datafile>
    <tns:location>/data/run1/file1.cbf</tns:location>
    <tns:size>12345</tns:size>
  </tns:datafile>
</tns:dataset>
"""

    data = raw_serialize_dataset_message(data, indent="  ")
    assert data.decode("utf-8") == expected


def test_investigation_message_xsd_only():
    data = {
        "experiment": "MX-123",
        "instrument": "ID30A-3",
        "startDate": "2026-03-08T09:18:49.678825+01:00",
        "investigationId": 4521,
    }

    expected = """<?xml version="1.0" encoding="UTF-8"?>
<tns:investigation xmlns:tns="http://www.esrf.fr/icat">
  <tns:experiment>MX-123</tns:experiment>
  <tns:instrument>ID30A-3</tns:instrument>
  <tns:startDate>2026-03-08T09:18:49.678825+01:00</tns:startDate>
  <tns:investigationId>4521</tns:investigationId>
</tns:investigation>
"""

    data = raw_serialize_investigation_message(data, indent="  ")
    assert data.decode("utf-8") == expected


def test_icat_serialize_warning(caplog):
    metadata = {}  # will cause a pydantic ValidationError about missing fields

    expected_message = _prepare_regex_pattern(
        r"""ValidationError: 6 validation errors for IcatDatasetParameters
title
  Field required [type=missing, input_value={}, input_type=dict]
    For further information visit __MISSING_URL__
proposal
  Field required [type=missing, input_value={}, input_type=dict]
    For further information visit __MISSING_URL__
folder_path
  Field required [type=missing, input_value={}, input_type=dict]
    For further information visit __MISSING_URL__
start_time
  Field required [type=missing, input_value={}, input_type=dict]
    For further information visit __MISSING_URL__
end_time
  Field required [type=missing, input_value={}, input_type=dict]
    For further information visit __MISSING_URL__
sample
  Field required [type=missing, input_value={}, input_type=dict]
    For further information visit __MISSING_URL__

Documentation on ICAT fields: __ICAT_DEF_URL__"""
    )

    with caplog.at_level(logging.WARNING, logger="pyicat_plus.client.serialize"):
        with pytest.warns(errors.IcatMetadataValidationWarning, match=expected_message):
            result = serialize_dataset_parameters(metadata)

    assert result == metadata
    messages = [
        record.message
        for record in caplog.records
        if record.levelno == logging.WARNING and "ValidationError" in record.message
    ]
    assert len(messages) == 1
    assert re.match(expected_message, messages[0])


def test_icat_serialize_error(caplog):
    metadata = {}  # will cause a pydantic ValidationError about missing fields

    with pytest.raises(errors.IcatMetadataValidationError):
        serialize_dataset_parameters(metadata, strict=True)


def _prepare_regex_pattern(template: str) -> str:
    pydantic_missing_url = r"https://errors\.pydantic\.dev/.*/v/missing"
    pydantic_value_error_url = r"https://errors\.pydantic\.dev/.*/v/value_error"
    icat_def_url = r"https://icat-esrf-definitions\.readthedocs\.io/.*/icat\.html"

    escaped = re.escape(template)

    escaped = escaped.replace("__MISSING_URL__", pydantic_missing_url)
    escaped = escaped.replace("__VALUE_ERROR_URL__", pydantic_value_error_url)
    escaped = escaped.replace("__ICAT_DEF_URL__", icat_def_url)

    return escaped



from dataclasses import dataclass
from dataclasses import field
from typing import List
from typing import Optional

from xsdata.models.datatype import XmlDateTime

__NAMESPACE__ = "http://www.esrf.fr/icat"


@dataclass
class Parameter:
    class Meta:
        name = "parameter"

    name: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.esrf.fr/icat",
            "required": True,
        },
    )
    value: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.esrf.fr/icat",
            "required": True,
        },
    )


@dataclass
class Datafile:
    class Meta:
        name = "datafile"

    location: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.esrf.fr/icat",
            "required": True,
        },
    )
    size: Optional[int] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.esrf.fr/icat",
        },
    )
    parameter: List[Parameter] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "http://www.esrf.fr/icat",
        },
    )


@dataclass
class Sample:
    class Meta:
        name = "sample"

    name: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.esrf.fr/icat",
            "required": True,
        },
    )
    type_value: Optional[str] = field(
        default=None,
        metadata={
            "name": "type",
            "type": "Element",
            "namespace": "http://www.esrf.fr/icat",
            "required": True,
        },
    )
    parameter: List[Parameter] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "http://www.esrf.fr/icat",
        },
    )


@dataclass
class Dataset:
    class Meta:
        name = "dataset"
        namespace = "http://www.esrf.fr/icat"

    dataset_id: Optional[int] = field(
        default=None,
        metadata={
            "name": "datasetId",
            "type": "Element",
        },
    )
    investigation: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "required": True,
        },
    )
    instrument: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "required": True,
        },
    )
    name: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "required": True,
        },
    )
    location: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "required": True,
        },
    )
    start_date: Optional[XmlDateTime] = field(
        default=None,
        metadata={
            "name": "startDate",
            "type": "Element",
            "required": True,
        },
    )
    end_date: Optional[XmlDateTime] = field(
        default=None,
        metadata={
            "name": "endDate",
            "type": "Element",
            "required": True,
        },
    )
    comment: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
        },
    )
    sample: Optional[Sample] = field(
        default=None,
        metadata={
            "type": "Element",
        },
    )
    datafile: List[Datafile] = field(
        default_factory=list,
        metadata={
            "type": "Element",
        },
    )
    parameter: List[Parameter] = field(
        default_factory=list,
        metadata={
            "type": "Element",
        },
    )
    complete: bool = field(
        default=True,
        metadata={
            "type": "Attribute",
        },
    )



from typing import Any
from typing import Dict
from typing import Tuple
from xml.etree import ElementTree

from xsdata.formats.dataclass.parsers import XmlParser
from xsdata.formats.dataclass.parsers.config import ParserConfig
from xsdata.formats.dataclass.serializers import DictEncoder
from xsdata.formats.dataclass.serializers import DictFactory

from ..models.xsd.dataset import Dataset
from ..models.xsd.investigation import Investigation


def deserialize_message(data: bytes) -> Tuple[str, Dict[str, Any]]:
    """
    :raises xml.etree.ElementTree.ParseError:
    """
    root = ElementTree.fromstring(data)
    root_tag = root.tag.split("}", 1)[-1]

    if root_tag == "dataset":
        cls = Dataset
    elif root_tag == "investigation":
        cls = Investigation
    else:
        raise ValueError(f"Unknown ICAT root element: {root.tag}")

    config = ParserConfig(fail_on_unknown_properties=False)
    deserializer = XmlParser(config=config)
    xsdata_model = deserializer.from_bytes(data, cls)

    encoder = DictEncoder(dict_factory=DictFactory.FILTER_NONE)
    model_as_dict = encoder.encode(xsdata_model)

    return root_tag, model_as_dict



from typing import Any
from typing import Dict

from .... import errors
from ...utils import constants
from .fallback import fallback_serialization
from .pydantic import pydantic_serialization


def serialize_dataset_parameters(
    parameters: Dict[str, Any], strict: bool = False
) -> Dict[str, str]:
    """In the XSD dataset schema the parameters are all strings.

    This method serializes parameter values with Pydantic validation and type coercion.

    :raises IcatMetadataSerializationError: If metadata serialization fails.
    :raises IcatMetadataValidationError: If ``strict=True`` and the metadata is invalid.
    :emits IcatMetadataValidationWarning: If ``strict=False`` and the metadata is invalid.
    :emits IcatMetadataDeprecatedWarning: If metadata is deprecated.
    """
    parsed_parameters, unknown_keys, error = pydantic_serialization(
        parameters, strict=strict
    )

    issues = []
    if error is not None:
        issues.append(f"{type(error).__name__}: {error}")
    if unknown_keys:
        issues.append(f"Unknown ICAT fields: {unknown_keys}")
    if issues:
        issues.append(
            f"Documentation on ICAT fields: https://icat-esrf-definitions.readthedocs.io/en/v{constants.definitions_release}/icat.html"
        )

        message = "\n\n".join(issues)
        if strict:
            raise errors.IcatMetadataValidationError(message) from error

        errors.warn(message, errors.IcatMetadataValidationWarning)

    if unknown_keys:
        unknown_parameters = {k: v for k, v in parameters.items() if k in unknown_keys}
        parsed_parameters.update(unknown_parameters)

    # Ensure all values are strings
    return fallback_serialization(parsed_parameters)



import json
import logging
import os
import socket
import threading
from enum import Enum
from json.decoder import JSONDecodeError
from typing import Optional
from xml.etree.ElementTree import ParseError

import stomp

from ..._icat_messaging.deserializers.xml import deserialize_message
from ...utils.log_utils import basic_config
from .activemq_rest_server import ICAT_QUEUES
from .icat_db import IcatDb

logger = logging.getLogger("STOMP SUBSCRIBER")

MessageType = Enum("MessageType", "investigation dataset archiving addfiles unknown")


class MyListener(stomp.ConnectionListener):
    def __init__(self, conn, icat_data_dir: Optional[str] = None):
        self.conn = conn
        self.s_out = None
        self.icatdb = IcatDb(icat_data_dir)
        super().__init__()

    def redirect_messages(self, port):
        if self.s_out is not None:
            self.s_out.close()
        self.s_out = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.s_out.connect(("localhost", port))
        logger.info(f"Redirect received messages to port {port}")

    def on_message(self, frame):
        message = frame.body
        logger.info("received message:\n %s", message)

        try:
            message_type, data = deserialize_message(message.encode("utf-8"))

            if message_type == "investigation":
                message_type = MessageType.investigation
            elif message_type == "dataset":
                message_type = MessageType.dataset
            else:
                message_type = None
            logger.info("Parsed XML message (%s):\n %s", message_type, data)

        except ParseError:
            try:
                data = json.loads(message)

                keys = set(data)
                if keys == {"datasetId", "type", "level", "message"}:
                    message_type = MessageType.archiving
                elif keys == {"datasetId"}:
                    message_type = MessageType.addfiles
                else:
                    message_type = MessageType.unknown

                logger.info("Parsed JSON message (%s):\n %s", message_type, data)
            except JSONDecodeError:
                data = None
                message_type = MessageType.unknown
                logger.info("Message parsing failed (%s):\n %s", message_type, data)

        # Only access specific destinations
        header = frame.headers
        icat_queues = ["/queue/" + q for q in ICAT_QUEUES]
        if header.get("destination") not in icat_queues:
            return

        # Only accept valid proposals
        if message_type in (message_type.investigation, message_type.dataset):
            if message_type is message_type.investigation:
                proposal = data["experiment"]
            else:
                proposal = data["investigation"]
                # Convert to backend format
                file_count = 0
                if os.path.exists(data["location"]):
                    file_count = len(os.listdir(data["location"]))
                data["parameter"].append({"name": "__fileCount", "value": file_count})
                data["parameters"] = data.pop("parameter")
            if proposal and "666" in proposal:
                logger.info(
                    "Do not register %s for invalid proposal '%s'",
                    message_type,
                    proposal,
                )
                return

        # Store data
        if message_type in (message_type.investigation, message_type.dataset):
            if message_type is message_type.investigation:
                self.icatdb.start_investigation(data)
            else:
                self.icatdb.store_dataset(data)

        if message_type == message_type.addfiles:
            dataset_id = data["datasetId"]
            with self.icatdb.update_dataset(dataset_id) as data:
                if data is None:
                    logger.error("datasetId %s does not exist", dataset_id)
                else:
                    file_count = 0
                    dirname = data["location"]
                    if os.path.exists(dirname):
                        file_count = len(os.listdir(dirname))
                    for parameter in data["parameters"]:
                        if parameter["name"] == "__fileCount":
                            logger.info(
                                "Update file count for %s to %d", dirname, file_count
                            )
                            parameter["value"] = file_count
                            break
                    else:
                        logger.info("Add file count for %s to %d", dirname, file_count)
                        data["parameters"].append(
                            {"name": "__fileCount", "value": file_count}
                        )

        # Notify that data is valid
        if self.s_out is not None:
            self.s_out.sendall(frame.body.encode() + b"\n")


def main(
    host=None, port=60001, queue=None, port_out=0, icat_data_dir: Optional[str] = None
):
    if not host:
        host = "localhost"
    if not queue:
        queue = "/queue/icatIngest"
    conn = stomp.Connection([(host, port)])
    # Listener will run in a different thread
    listener = MyListener(conn, icat_data_dir)
    conn.set_listener("", listener)
    conn.connect("guest", "guest", wait=True)
    conn.subscribe(destination=queue, id=1, ack="auto")
    logger.info(f"subscribed to {queue} on STOMP {host}:{port}")
    if port_out:
        listener.redirect_messages(port_out)
        listener.s_out.sendall(b"LISTENING\n")
    logger.info("CTRL-C to stop")
    try:
        threading.Event().wait()
    finally:
        logger.info("Exit.")


if __name__ == "__main__":
    import argparse

    basic_config(
        logger=logger,
        level=logging.DEBUG,
        format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
    )

    parser = argparse.ArgumentParser(
        description="STOMP client which subscribes to a STOMP queue and redirect its output to a socket"
    )
    parser.add_argument(
        "--host", default="localhost", type=str, help="STOMP server host"
    )
    parser.add_argument("--port", default=60001, type=int, help="STOMP server port")
    parser.add_argument(
        "--queue", default="/queue/icatIngest", type=str, help="STOMP queue"
    )
    parser.add_argument("--port_out", default=0, type=int, help="output socket")
    parser.add_argument(
        "--icat_data_dir", default=None, type=str, help="Dataset directory"
    )
    args = parser.parse_args()

    main(
        host=args.host,
        port=args.port,
        port_out=args.port_out,
        queue=args.queue,
        icat_data_dir=args.icat_data_dir,
    )



import datetime

from ..._icat_messaging.serializers.investigation import (
    raw_serialize_investigation_message,
)
from ..._icat_messaging.serializers.investigation import serialize_investigation_message
from ..._icat_messaging.utils import datetime_utils


def test_investigation_message():
    proposal = "MX-123"
    beamline = "ID30A-3"
    start_datetime = datetime.datetime.now().astimezone()

    expected = f"""<?xml version="1.0" encoding="UTF-8"?>
<tns:investigation xmlns:tns="http://www.esrf.fr/icat">
  <tns:experiment>MX-123</tns:experiment>
  <tns:instrument>ID30A-3</tns:instrument>
  <tns:startDate>{datetime_utils.with_timezone_zulu_format(start_datetime)}</tns:startDate>
</tns:investigation>
"""

    data = serialize_investigation_message(
        beamline=beamline, proposal=proposal, start_datetime=start_datetime, indent="  "
    )
    assert data.decode("utf-8") == expected


def test_investigation_message_xsd_only():
    data = {
        "experiment": "MX-123",
        "instrument": "ID30A-3",
        "startDate": "2026-03-08T09:18:49.678825+01:00",
        "investigationId": 4521,
    }

    expected = """<?xml version="1.0" encoding="UTF-8"?>
<tns:investigation xmlns:tns="http://www.esrf.fr/icat">
  <tns:experiment>MX-123</tns:experiment>
  <tns:instrument>ID30A-3</tns:instrument>
  <tns:startDate>2026-03-08T09:18:49.678825+01:00</tns:startDate>
  <tns:investigationId>4521</tns:investigationId>
</tns:investigation>
"""

    data = raw_serialize_investigation_message(data, indent="  ")
    assert data.decode("utf-8") == expected



import json
import logging
import os
from contextlib import contextmanager
from glob import glob
from typing import Any
from typing import Dict
from typing import Generator
from typing import List
from typing import Optional

from xsdata.models.datatype import XmlDateTime

from ..._icat_messaging.utils import datetime_utils

logger = logging.getLogger("STOMP SUBSCRIBER")


class IcatDb:
    def __init__(self, root_dir: Optional[str] = None):
        if root_dir is None:
            root_dir = "."
        if root_dir:
            os.makedirs(root_dir, exist_ok=True)
        self._root_dir = root_dir
        self._investigations = os.path.join(root_dir, "investigations.json")

    def start_investigation(self, investigation: dict) -> None:
        investigation["instrument"] = {"name": investigation["instrument"]}
        investigation["proposal"] = investigation["experiment"]
        _add_table(self._investigations, investigation)

    def store_dataset(self, dataset: dict) -> None:
        investigation = _find_data(self._investigations, dataset["startDate"])
        if investigation is None:
            investigation_id = _get_investigation_id(dataset)
            investigation = _find_investigation(self._investigations, investigation_id)
            if investigation is None:
                logger.error(
                    "Dataset not stored because no investigation found: %s", dataset
                )
                return
        filename = os.path.join(self._root_dir, f"datasets{investigation['id']}.json")
        dataset["id"] = self._get_next_dataset_id()
        _add_table(filename, dataset)

    @contextmanager
    def update_dataset(self, dataset_id: int) -> Generator[dict, None, None]:
        for filename in glob(os.path.join(self._root_dir, "datasets*.json")):
            data = _get_row(filename, dataset_id)
            if data is None:
                continue
            yield data
            _edit_table(filename, data)
            break
        else:
            yield

    def get_all_investigations(self) -> List[Dict]:
        return _read_table(self._investigations)

    def get_investigations(self, instrument: str, experiment: str) -> List[Dict]:
        return [
            investigation
            for investigation in _read_table(self._investigations)
            if experiment == investigation["experiment"]
            and instrument == investigation["instrument"]["name"]
        ]

    def get_datasets(self, investigation_id) -> List[Dict]:
        filename = os.path.join(self._root_dir, f"datasets{investigation_id}.json")
        return _read_table(filename)

    def _get_next_dataset_id(self) -> int:
        max_ids = [
            _get_max_id(filename)
            for filename in glob(os.path.join(self._root_dir, "datasets*.json"))
        ]
        if max_ids:
            return max(max_ids) + 1
        return 0


def _read_table(filename: str) -> List[dict]:
    if not os.path.isfile(filename):
        return list()
    with open(filename, "r") as f:
        return json.load(f)


def _write_table(filename: str, data: dict) -> Dict:
    with open(filename, "w") as f:
        json.dump(_serialize_icat_db(data), f)
    logger.info("Metadata in %s updated", filename)


def _serialize_icat_db(obj: Any) -> Any:
    """
    Convert objects that are not JSON serializable:
    - XmlDateTime -> ISO 8601 string
    - dataclass -> dict (recursively)
    - list/tuple -> recursively
    """
    if isinstance(obj, XmlDateTime):
        return datetime_utils.with_timezone_offset_format(obj)
    elif isinstance(obj, (list, tuple)):
        return [_serialize_icat_db(v) for v in obj]
    elif isinstance(obj, dict):
        return {k: _serialize_icat_db(v) for k, v in obj.items()}
    else:
        return obj


def _add_table(filename: str, data: dict) -> None:
    table = _read_table(filename)
    if "id" not in data:
        if table:
            data_id = max(row["id"] for row in table) + 1
        else:
            data_id = 0
        data = dict(data)
        data["id"] = data_id
    table.append(data)
    logger.info("Add metadata to %s: %s", filename, data)
    _write_table(filename, table)


def _edit_table(filename: str, data: dict) -> None:
    table = _read_table(filename)
    for row in table:
        if row["id"] == data["id"]:
            logger.info("Edit metadata of %s: %s", filename, data)
            row.update(data)
            _write_table(filename, table)
            break
    else:
        _add_table(filename, data)


def _get_row(filename: str, dataset_id: int) -> Optional[dict]:
    table = _read_table(filename)
    for row in table:
        if row["id"] == dataset_id:
            return row


def _get_max_id(filename: str) -> int:
    table = _read_table(filename)
    if table:
        return max(row["id"] for row in table)
    return -1


def _find_data(filename: str, date: str) -> Optional[Dict]:
    date = datetime_utils.deserialize_datetime(date)
    for data in _read_table(filename):
        start_date = datetime_utils.deserialize_datetime(data["startDate"])
        end_date = data.get("endDate")
        if end_date is None:
            # infinite timeslot
            inside_timeslot = date >= start_date
        else:
            # finite timeslot
            end_date = datetime_utils.deserialize_datetime(data["endDate"])
            inside_timeslot = date >= start_date and date <= end_date
        if inside_timeslot:
            return data


def _get_investigation_id(dataset: dict) -> Optional[str]:
    for param in dataset.get("parameter", []):
        if param["name"] == "investigation_id":
            return param["value"]
    return None


def _find_investigation(filename: str, investigation_id: str) -> Optional[Dict]:
    for data in _read_table(filename):
        if _get_investigation_id(data) == investigation_id:
            return data
